/*	$Id: PaletteGik.h 2008/04/07 17:30:13 mallet $ */

/*
 * Copyright (c) 2007 LAAS/CNRS                       --  Mon Apr 16 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef PALETTE_GIK_H
#define PALETTE_GIK_H

#include <string.h>
#include <portLib.h>

#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include "robotDynamics/jrlRobotDynamicsObjectConstructor.h"
#include "dynamicsJRLJapan/HumanoidDynamicMultiBody.h"

#include "gikTask/jrlGikObjectConstructor.h"

#include "constraints/hppGikPlaneConstraint.h"
#include "constraints/hppGikPositionConstraint.h"
#include "constraints/hppGikRotationConstraint.h"
#include "constraints/hppGikTransformationConstraint.h"
#include "constraints/hppGikParallelConstraint.h"
#include "constraints/hppGikGazeConstraint.h"
#include "constraints/hppGikComConstraint.h"
#include "core/hppGikSolver.h"

#include "tasks/hppGikRobotTask.h"

#include "netshring.h"

#include "gikStruct.h"

class PaletteGik;

extern netshring_id_t shtarget;		/**< shared trajectory buffer */
extern CjrlHumanoidDynamicRobot *robot; /**< the robot model */
extern PaletteGik *paletteGik;

STATUS	worldCoordinates(VEC& inCoords, vector3d& outCoords);

/** Interface to a set of ChppGikRobotTasks */
class PaletteGik
{

public:
    CjrlGikObjectConstructor<
    ChppGikPlaneConstraint,
    ChppGikParallelConstraint,
    ChppGikRotationConstraint,
    ChppGikPositionConstraint,
    ChppGikTransformationConstraint,
    ChppGikPointingConstraint,
    ChppGikGazeConstraint,
    ChppGikComConstraint> gikFactory;

    ChppGikStandingRobot* attStandingRobot;

    /**
     \brief Constructor
    */
    PaletteGik(const char *path, const char *model);

    /**
    \brief Destructor
    */
    ~PaletteGik();

    /**
    \brief Reset
    */
    void  reset();

    /**
    \name Motion planners
    {@
    */
    bool reach(bool taskIsForRightHand, bool doOrientation, bool doLook,
               bool steppingEnabled, const vector3d& targetPoint,
               const vector3d& targetOrientation, const vector3d& gazeTarget);

    bool grasp(bool taskIsForRightHand, const vector3d &targetPoint,
               double clench);

    bool grasp_via(bool taskIsForRightHand, const vector3d &viaPoint,
                   const vector3d &targetPoint, double clench);

    bool release(bool taskIsForRightHand, const vector3d &targetPoint,
                 double clench);

    bool release_via(bool taskIsForRightHand, const vector3d &viaPoint,
                     const vector3d &targetPoint, double clench);

    bool handGrasp(bool taskIsForRightHand, double valTighten);

    bool lookat(const vector3d& coordinates);
    bool halfsitting();
    bool stepback();


    bool sittingReach(bool taskIsForRightHand, bool doOrientation, bool doLook,
                      const vector3d& targetPoint,
                      const vector3d& targetOrientation, const vector3d& gazeTarget);
    bool sittingLookat(const vector3d& coordinates);

    bool sittingHandGrasp(bool taskIsForRightHand, double valTighten);

    bool sittingManipulation(bool isForGrasp,bool taskIsForRightHand, const std::vector<vector3d>& inPickingPath, const std::vector<vector3d>& inReturningPath, double clench);

    void sittingSetConfiguration(const vectorN& inConfiguration);
    /**
    @}
    */

    void apply();
    void dumpLastMotion();
    void setConfiguration(vectorN& inConfiguration);

    bool firstSample(vectorN &config, vector3d &ZMPwstObs, double &timeRef);
    bool nextSample(vectorN &config, vector3d &ZMPwstObs, double &timeRef);

private:

    bool solveTaskStack(std::vector<ChppGikVectorizableConstraint*>& inTasks);

    bool elementLookat(const vector3d& targetPoint);
    bool elementReach(bool taskIsForRightHand, bool doOrientation, bool doLook,
                      const vector3d& inReachTarget,
                      const vector3d& inTargetOrientation,
                      const vector3d& inGazeTarget);
    bool elementClench(bool taskIsForRightHand, double clench);

    bool solveConstraintStack(std::vector<CjrlGikStateConstraint*>& inConstraints, const std::vector<double>& inDamp);

    bool solveStack();


    void createStandingRobot(const char *ipath, const char *imodel);

    CjrlGikObjectConstructor<ChppGikPlaneConstraint, ChppGikParallelConstraint, ChppGikRotationConstraint, ChppGikPositionConstraint, ChppGikTransformationConstraint, ChppGikPointingConstraint, ChppGikGazeConstraint, ChppGikComConstraint> attGikFactory;

    double attSamplingPeriod;
    double attLastSampleTime;

    void cropMotion(ChppGikRobotTask* robotTask);
    void cropGik();
    void crop(ChppRobotMotion& inSource, ChppRobotMotion& outContainer);

    ChppRobotMotion* attLastMotion,*attGikMotion;

    string attHRP2MotionName;

    vectorN attCurrentStartConfig;
    vectorN attNextStartConfig, attBackupVelocity, attBackupConfig;

    ChppGikSolver *attGikSolver;

    double attSittingTasksDuration;

};

#endif
