/*	$Id: gikBoardCodels.cc 2008/10/09 10:12:30 mallet $ */

/*
 * Copyright (c) 2007 Oussama Kanoun                  --  Mon Apr 16 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: gikBoardCodels.cc 2008/10/09 10:12:30 mallet $";

#include <err.h>
#include <portLib.h>

#include "server/gikHeader.h"
#include "genom-openhrp/genom-hrp2.h"

#include "PaletteGik.h"


/* --- local data ---------------------------------------------------- */

netshring_id_t shtarget;		/**< shared trajectory buffer */

CjrlHumanoidDynamicRobot *robot;	/**< the robot model */
PaletteGik *paletteGik;


/* --- BoardStart ---------------------------------------------------- */

/** Initialization codel (fIDS, ...)
 *
 * \return OK or ERROR
 */

STATUS
gikBoardStart(int *report)
{
    /* create shared trajectory buffer */
    shtarget = netshring_create(GIK_TARGET_CONFIG_NAME);
    if (!shtarget)
    {
        warnx("cannot create shared buffer");
        return ERROR;
    }

    return OK;
}


/* --- BoardEnd ------------------------------------------------------ */

/** Termination codel
 *
 * \returns OK or ERROR
 */

STATUS
gikBoardEnd(void)
{
    /* delete shared trajectory buffer */
    if (shtarget)
    {
        netshring_delete(shtarget);
        shtarget = NULL;
    }

    if (paletteGik)
        delete paletteGik;
    return OK;
}


/* --- Init ---------------------------------------------------------- */

/** gikInitMain  -  codel EXEC of Init
 *
 * Reports:      OK
 *              S_gik_NO_MODEL
 * \returns EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
gikInitMain ( GikString *model, int *report )
{
    char *delim = strrchr ( model->name, '/' );

    if ( !delim )
        paletteGik = new PaletteGik ( "./", model->name );
    else
    {
        * ( delim++ ) = 0;
        paletteGik = new PaletteGik ( model->name, delim );
    }

    return ETHER;
}



/*------------------------------------------------------------------------
 * Look
 *
 * Description:
 *
 * Reports:      OK
 */

/* LookMain  -  codel EXEC of Look
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LookMain ( VEC *look_input, BOOLEAN *outSuccess, int *report )
{
    vector3d p;

    worldCoordinates ( *look_input,p );

    bool solved = paletteGik->lookat ( p );

    if ( solved )
        *outSuccess = True;
    else
        *outSuccess = False;
    return ETHER;
}

/*------------------------------------------------------------------------
 * Sitting Look
 *
 * Description:
 *
 * Reports:      OK
 */

/* SittingLookMain  -  codel EXEC of Look
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SittingLookMain ( VEC *look_input, BOOLEAN *outSuccess, int *report )
{
    vector3d p;

    worldCoordinates ( *look_input,p );

    bool solved = paletteGik->sittingLookat ( p );

    if ( solved )
        *outSuccess = True;
    else
        *outSuccess = False;
    return ETHER;
}

/*------------------------------------------------------------------------
 * ABSLook
 *
 * Description:
 *
 * Reports:      OK
 */

/* ABSLookMain  -  codel EXEC of Look
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ABSLookMain ( VEC *look_input, BOOLEAN *outSuccess, int *report )
{
    vector3d p;

    p[0] = look_input->x;
    p[1] = look_input->y;
    p[2] = look_input->z;

    bool solved = paletteGik->lookat ( p );

    if ( solved )
        *outSuccess = True;
    else
        *outSuccess = False;
    return ETHER;
}


/* --- Reach --------------------------------------------------------- */

/** Place one hand at the given position
 *
 *
 * \return	OK
 *		S_gik_NO_SOLUTION
 */

/* ReachMain  -  codel EXEC of Reach
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikReachMain(GikReachData *reach_input, int *report)
{
    vector3d p,o,g;

    bool isForRightHand = (reach_input->hand == GIK_RIGHT_HAND);
    bool enableStepping = (reach_input->enableStepping == True);
    bool doOrientation = (reach_input->orient_job == True);
    bool doLook = (reach_input->look_job == True);
    bool solved;

    worldCoordinates(reach_input->position, p);
    worldCoordinates(reach_input->axis, o);
    worldCoordinates(reach_input->gazepoint, g);

    solved = paletteGik->reach(
                 isForRightHand, doOrientation, doLook, enableStepping, p, o, g);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}

/* --- Sitting Reach --------------------------------------------------------- */

/** Place one hand at the given position
 *
 *
 * \return	OK
 *		S_gik_NO_SOLUTION
 */

/* SittingReachMain  -  codel EXEC of Reach
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SittingReachMain(GikSittingReachData *sitting_reach_input, int *report)
{
    vector3d p,o,g;

    bool isForRightHand = (sitting_reach_input->hand == GIK_RIGHT_HAND);
    bool doOrientation = (sitting_reach_input->orient_job == True);
    bool doLook = (sitting_reach_input->look_job == True);
    bool solved;

    worldCoordinates(sitting_reach_input->position, p);
    worldCoordinates(sitting_reach_input->axis, o);
    worldCoordinates(sitting_reach_input->gazepoint, g);

    solved = paletteGik->sittingReach(isForRightHand, doOrientation, doLook, p, o, g);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}

/* --- Grasp --------------------------------------------------------- */

/** Generate a simple grasping motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikGraspMain(GikGraspData *grasp_input, int *report)
{
    bool solved;
    vector3d p;

    bool isForRightHand = (grasp_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(grasp_input->position, p);

    solved = paletteGik->grasp(isForRightHand, p, grasp_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}


/* --- GraspVia ------------------------------------------------------ */

/** Generate a simple grasping motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikGraspViaMain(GikGraspViaData *grasp_via_input, int *report)
{
    bool solved;
    vector3d v, p;

    bool isForRightHand = (grasp_via_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(grasp_via_input->via, v);
    worldCoordinates(grasp_via_input->position, p);

    solved = paletteGik->grasp_via(
                 isForRightHand, v, p, grasp_via_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}



/* --- SittingGraspVia ------------------------------------------------------ */

/** Generate a simple grasping motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikSittingGraspViaMain(GikSittingViaData *sitting_grasp_via_input, int *report)
{
    bool solved;
    vector3d v, p, r;

    bool isForRightHand = (sitting_grasp_via_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(sitting_grasp_via_input->via, v);
    worldCoordinates(sitting_grasp_via_input->position, p);
    worldCoordinates(sitting_grasp_via_input->returnPosition, r);

    std::vector<vector3d> PickPath,RetPath;
    if ((p[0] == v[0]) && (p[1] == v[1]) && (p[2] == v[2]))
        PickPath.push_back(p);
    else
    {
        PickPath.push_back(v);
        PickPath.push_back(p);
    }

    if ((p[0] != r[0]) || (p[1] != r[1]) || (p[2] != r[2]))
        RetPath.push_back(r);

    solved = paletteGik->sittingManipulation(true,isForRightHand,PickPath,RetPath, sitting_grasp_via_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}

/* --- SittingReleaseVia ------------------------------------------------------ */

/** Generate a simple releasing motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikSittingReleaseViaMain(GikSittingViaData *sitting_release_via_input, int *report)
{
    bool solved;
    vector3d v, p, r;

    bool isForRightHand = (sitting_release_via_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(sitting_release_via_input->via, v);
    worldCoordinates(sitting_release_via_input->position, p);
    worldCoordinates(sitting_release_via_input->returnPosition, r);

    std::vector<vector3d> PickPath,RetPath;
    if ((p[0] == v[0]) && (p[1] == v[1]) && (p[2] == v[2]))
        PickPath.push_back(p);
    else
    {
        PickPath.push_back(v);
        PickPath.push_back(p);
    }

    if ((p[0] != r[0]) || (p[1] != r[1]) || (p[2] != r[2]))
        RetPath.push_back(r);

    solved = paletteGik->sittingManipulation(false,isForRightHand,PickPath,RetPath, sitting_release_via_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}


/* --- Release ------------------------------------------------------- */

/** Generate a simple release motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikReleaseMain(GikGraspData *grasp_input, int *report)
{
    bool solved;
    vector3d p;

    bool isForRightHand = (grasp_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(grasp_input->position, p);

    solved = paletteGik->release(isForRightHand, p, grasp_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}


/* --- ReleaseVia ---------------------------------------------------- */

/** Generate a simple release motion at the given 3D point
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

ACTIVITY_EVENT
gikReleaseViaMain(GikGraspViaData *grasp_via_input, int *report)
{
    bool solved;
    vector3d v, p;

    bool isForRightHand = (grasp_via_input->hand == GIK_RIGHT_HAND);

    worldCoordinates(grasp_via_input->via, v);
    worldCoordinates(grasp_via_input->position, p);

    solved = paletteGik->release_via(
                 isForRightHand, v, p, grasp_via_input->clench);

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}


/* --- HalfSitting --------------------------------------------------- */

/** Go to the (initial) half-sitting stance
 *
 * \return	OK,
 *		S_gik_NO_SOLUTION
 */

/* gikHalfSittingMain  -  codel EXEC of Halfsitting
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
gikHalfSittingMain(int *report)
{
    bool solved = paletteGik->halfsitting();

    if (!solved)
        *report = S_gik_NO_SOLUTION;
    return ETHER;
}


/*------------------------------------------------------------------------
 * HandClench
 *
 * Description:
 *
 * Reports:      OK
 */

/* HandClenchMain  -  codel EXEC of HandClench
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
HandClenchMain ( CLENCHDATA *handclench_input, BOOLEAN *outSuccess, int *report )
{
    bool isForRightHand = ( handclench_input->right == 1 );
    double clench = handclench_input->clench;
    bool solved = paletteGik->handGrasp ( isForRightHand, clench );
    if ( solved )
        *outSuccess = True;
    else
        *outSuccess = False;
    return ETHER;
}

/*------------------------------------------------------------------------
 * SittingHandClench
 *
 * Description:
 *
 * Reports:      OK
 */

/* SittingHandClenchMain  -  codel EXEC of HandClench
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SittingHandClenchMain ( CLENCHDATA *handclench_input, BOOLEAN *outSuccess, int *report )
{
    bool isForRightHand = ( handclench_input->right == 1 );
    double clench = handclench_input->clench;
    bool solved = paletteGik->sittingHandGrasp ( isForRightHand, clench );
    if ( solved )
        *outSuccess = True;
    else
        *outSuccess = False;
    return ETHER;
}


/* --- AcceptSolution ------------------------------------------------ */

/** AcceptSolutionMain  -  codel EXEC of AcceptSolution
 *
 * Reports:      OK
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
gikAcceptSolutionMain ( int *report )
{
    paletteGik->apply();
    return ETHER;
}


/*------------------------------------------------------------------------
 * DumpLastMotion
 *
 * Description:
 *
 * Reports:      OK
 */

/* DumpLastMotionMain  -  codel EXEC of DumpLastMotion
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
DumpLastMotionMain ( int *report )
{
    paletteGik->dumpLastMotion();
    return ETHER;
}


/* --- GetConfig ----------------------------------------------------- */

/** gikGetConfigMain  -  codel EXEC of GetConfig
 *
 * Reports:      OK
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
gikGetConfigMain ( int *report )
{
    vectorN motionConfig;
    ghrp2_config_t cfg;
    POSTER_ID p;
    STATUS s;
    int i;

    s = posterFind ( "hrp2State", &p );
    if ( s != OK )
    {
        *report = S_gik_NO_MODEL;
        return ETHER;
    }

    if (posterRead(p, sizeof(cfg), &cfg, sizeof(cfg)) != sizeof(cfg))
    {
        *report = S_gik_NO_MODEL;
        return ETHER;
    }

    motionConfig.resize ( 46 );
    motionConfig[0] = 0;
    motionConfig[1] = 0;
    motionConfig[2] = 0;
    motionConfig[3] = cfg.waistRpy[0];
    motionConfig[4] = cfg.waistRpy[1];
    motionConfig[5] = cfg.waistRpy[2];
    for ( i=0; i<40; i++ )
    {
        motionConfig[i+6] = cfg.angles[i];
    }

    paletteGik->setConfiguration ( motionConfig );
    return ETHER;
}

/* --- GetConfig ----------------------------------------------------- */

/** gikGetConfigMain  -  codel EXEC of GetConfig
 *
 * Reports:      OK
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
SittingGetConfigMain ( int *report )
{
    vectorN motionConfig;
    ghrp2_config_t cfg;
    POSTER_ID p;
    STATUS s;
    int i;

    s = posterFind ( "hrp2State", &p );
    if ( s != OK )
    {
        *report = S_gik_NO_MODEL;
        return ETHER;
    }

    if (posterRead(p, sizeof(cfg), &cfg, sizeof(cfg)) != sizeof(cfg))
    {
        *report = S_gik_NO_MODEL;
        return ETHER;
    }

    motionConfig.resize ( 46 );
    motionConfig[0] = 0;
    motionConfig[1] = 0;
    motionConfig[2] = 0;
    motionConfig[3] = 0;
    motionConfig[4] = 0;
    motionConfig[5] = 0;
    for ( i=0; i<40; i++ )
    {
        motionConfig[i+6] = cfg.angles[i];
    }

    paletteGik->sittingSetConfiguration ( motionConfig );
    return ETHER;
}

/* --- ExecuteMotion ------------------------------------------------- */

/** ExecutionMotionMain  -  codel EXEC of ExecuteMotion
 *
 * Reports	OK,
 *		S_gik_NO_MOTION,
 *		S_gik_CANNOT_PUSH
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
gikExecuteMotionMain(int *report)
{
    vectorN motionConfig;
    vector3d motionZMPwstObs;
    double motionTimeRef;
    ghrp2_config_t cfg;
    size_t s;
    int i, j;

    j = 0;
    for (bool ok = paletteGik->firstSample(
                       motionConfig, motionZMPwstObs, motionTimeRef);
            ok;
            ok = paletteGik->nextSample(
                     motionConfig, motionZMPwstObs, motionTimeRef))
    {

        cfg.time = motionTimeRef;
        for (i = 0; i<GHRP2_MAXDOF; i++)
        {
            cfg.angles[i] = motionConfig[i+6];
        }
        cfg.zmp[0] = motionZMPwstObs[0];
        cfg.zmp[1] = motionZMPwstObs[1];
        cfg.zmp[2] = motionZMPwstObs[2];

        cfg.waistRpy[0] = motionConfig[3];
        cfg.waistRpy[1] = motionConfig[4];
        cfg.waistRpy[2] = motionConfig[5];

        s = netshring_push(shtarget, &cfg, sizeof(cfg));
        if (s < 0 || s != sizeof(cfg))
        {
            *report = S_gik_CANNOT_PUSH;
            return ETHER;
        }
    }

    printf("wrote config up to time %g\n", cfg.time);

    cfg.time = -1.0;
    s = netshring_push(shtarget, &cfg, sizeof(cfg));
    if (s < 0 || s != sizeof(cfg))
    {
        *report = S_gik_CANNOT_PUSH;
        return ETHER;
    }

    return ETHER;
}


/* --- worldCoordinates ---------------------------------------------- */

/** Transform world coordinates into waist coordinates
 *
 */

STATUS
worldCoordinates(VEC& inCoords, vector3d& outCoords)
{
    matrix4d m = robot->waist()->currentTransformation();
    vector3d p;
    p[0] = inCoords.x;
    p[1] = inCoords.y;
    p[2] = inCoords.z;
    outCoords = m * p;
    return OK;
}
